"""Main module supporting repeated polling."""

import os
import sys
import logging

from datetime import datetime, timedelta, timezone
import threading

from omnik.ha_logger import hybridlogger

logging.basicConfig(stream=sys.stdout, level=os.environ.get("LOGLEVEL", logging.INFO))

__version__ = "1.15.2"

logger = logging.getLogger(__name__)

LOGLEVEL = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
}


class RepeatedJob(object):
    """Add timer support."""

    def __init__(self, c, datalogger, hass_api, *args, **kwargs):
        if c.get("default", "debug", fallback=False):
            logger.setLevel(logging.DEBUG)
        if c.get("default", "loglevel"):
            logger.setLevel(LOGLEVEL.get(c.get("default", "loglevel")))

        self.logger = logger
        self.hass_api = hass_api
        self.datalogger = datalogger
        self.client = datalogger.client
        self.args = args
        self.kwargs = kwargs
        # Initialize retry counter
        self.is_running = False
        self.last_update_time = None
        self.new_report_expected_at = None
        self.use_timer = datalogger.client.use_timer
        if self.use_timer:
            self._timer = None
            self.interval = int(c.get("default", "interval", fallback=360))
            self.half_interval = self.interval / 2
            self.retries = 0
            # Trigger the RepeatedJob using a short timer (1s) for the first time so the initialization can finish
            self.calculated_interval = 1
        else:
            self.semaphore = self.client.semaphore
            self.msgevent = self.client.msgevent
        self.start()

    def function_thread(self):
        """Excecute the client code."""
        return self.function_thread

    # This handler function is fired when the timer has reached the interval
    def _run(self):
        self.is_running = False
        # The function calls DataLogger.process()
        self.last_update_time = self.datalogger.process(*self.args, **self.kwargs)
        # Calculate the new timer interval
        if self.last_update_time:
            # Reset retry counter
            self.retries = 0
            if self.last_update_time <= datetime.now(timezone.utc):
                # If last report time + 2x interval is less than the current time then increase
                self.new_report_expected_at = self.last_update_time + timedelta(
                    seconds=self.interval
                )
                # Check if we have at least 60 seconds for the next cycle
                if self.new_report_expected_at + timedelta(seconds=-10) < datetime.now(
                    timezone.utc
                ):
                    # No recent update or missing update: wait {interval}/5 from now()
                    self.new_report_expected_at = datetime.now(
                        timezone.utc
                    ) + timedelta(seconds=self.interval / 5)
            else:
                # Skipping dark period
                self.new_report_expected_at = self.last_update_time
            self.calculated_interval = (
                self.new_report_expected_at - datetime.now(timezone.utc)
            ).seconds
        else:
            # An error occured calculate retry interval
            retry_interval = self.half_interval
            i = self.retries
            while i > 0:
                # Double retry interval to avoid to much traffic
                retry_interval *= 2
                i -= 1
            # Increment retry counter maximal interval between retries is half_interval * 2 * 2 * 2 = 4 intervals
            if self.retries < 3:
                self.retries += 1
            # Calculate new report time
            self.new_report_expected_at = datetime.now(timezone.utc) + timedelta(
                seconds=retry_interval
            )
            self.calculated_interval = (
                self.new_report_expected_at - datetime.now(timezone.utc)
            ).seconds
        # Make sure we have at least 15 seconds on the time to prevent a deadlocked timer loop
        if self.calculated_interval < 15:
            self.calculated_interval = 15
        hybridlogger.ha_log(
            self.logger,
            self.hass_api,
            "DEBUG",
            f"new poll in {self.calculated_interval} seconds at {self.new_report_expected_at.isoformat()}.",
        )
        self.start()

    # This function actual starts the timer
    def start(self):
        """Start a new timer."""
        if self.use_timer:
            # starting actual timer
            if not self.is_running:
                self._timer = threading.Timer(self.calculated_interval, self._run)
                self._timer.daemon = True
                self._timer.start()
                self.is_running = True
        else:
            # use a listing thread to process
            self.listenthread = threading.Thread(target=self._listen_to_events)
            self.listenthread.start()

    def stop(self):
        """Cancel the timer."""
        if self.use_timer:
            self._timer.cancel()
        else:
            # exit event message loop
            self.msgevent.set()
        self.is_running = False

    def _listen_to_events(self):
        self.is_running = True
        while self.is_running:
            self.last_update_time = self.datalogger.process(*self.args, **self.kwargs)



from influxdb_client import InfluxDBClient
from influxdb_client.client.write_api import SYNCHRONOUS
from influxdb_client.client.exceptions import InfluxDBError
import requests
from omnik.ha_logger import hybridlogger

from omnik.plugin_output import Plugin
import threading


class influxdb(Plugin):
    def __init__(self):
        super().__init__()
        self.auth = None
        self.client = None
        self.name = "influxdb"
        self.description = "Write output to InfluxDB"
        # common settings
        self.host = self.config.get("output.influxdb", "host", fallback="localhost")
        self.port = self.config.get("output.influxdb", "port", fallback="8086")
        self.ssl = self.config.getboolean("output.influxdb", "ssl", fallback=False)
        self.verify_ssl = self.config.getboolean(
            "output.influxdb", "verify_ssl", fallback=True
        )
        self.ssl_ca_cert = self.config.get("output.influxdb", "ssl_ca_cert")
        self.path = self.config.get("output.influxdb", "path")

        # Influx version 1 settings
        self.database = self.config.get(
            "output.influxdb", "database", fallback="omnikdatalogger"
        )

        self.headers = {"Content-type": "application/x-www-form-urlencoded"}
        # Add JWT authentication token
        if self.config.has_option("output.influxdb", "jwt_token"):
            self.headers[
                "Authorization"
            ] = f"Bearer {self.config.get('output.influxdb', 'jwt_token')}"
        elif self.config.has_option(
            "output.influxdb", "username"
        ) and self.config.has_option("output.influxdb", "password"):
            self.auth = requests.auth.HTTPBasicAuth(
                self.config.get("output.influxdb", "username"),
                self.config.get("output.influxdb", "password"),
            )
        # Influx vs 2 settings
        self.org = self.config.get("output.influxdb", "org")
        self.bucket = self.config.get("output.influxdb", "bucket")
        self.token = self.config.get("output.influxdb", "token")
        self.auth_v2 = False
        if self.org and self.bucket and self.token:
            self.auth_v2 = True
            # Log warning if that v1 auth will be ignored
            if self.auth:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "WARNING",
                    "Ignoring InfluxDB v1 settings database, username, password and jwt_token",
                )
            protocol = "https" if self.ssl else "http"
            path = f"/{self.path}" if self.path else ""
            url = f"{protocol}://{self.host}:{self.port}{path}"
            self.client = InfluxDBClient(
                url=url,
                token=self.token,
                org=self.org,
                verify_ssl=self.verify_ssl,
                ssl_ca_cert=self.ssl_ca_cert,
            )
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                "InfluxDB v2 client initialized",
            )
        else:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                "InfluxDB v1 client initialized",
            )
            if not self.auth:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "WARNING",
                    "Authentication was set or incomplete!",
                )

        self.timestamp_field = {}
        for field in self.config.data_field_config:
            if not self.config.data_field_config[field]["dev_cla"] == "timestamp":
                continue
            self.timestamp_field[self.config.data_field_config[field]["asset"]] = field
        self.access = threading.Condition(threading.Lock())

    def _get_temperature(self, values):
        if self.config.getboolean("output.influxdb", "use_temperature", fallback=False):
            weather = self.get_weather()
            values["temperature"] = str(weather["main"]["temp"])
            values["timestamp_ow"] = weather["main"]["dt"]

    def _get_attributes(self, values, asset_class):
        attributes = {}
        for att in self.config.attributes["asset"][asset_class]:
            if att in self.config.data_field_config:
                if self.config.data_field_config[att]["dev_cla"] == "timestamp":
                    continue
            attributes[att] = values[att]
        return attributes.copy()

    def _get_tags(self, field, attributes, asset_class):
        tags = attributes.copy()
        tags["asset"] = asset_class
        tags["entity"] = field
        tags["name"] = str(self.config.data_field_config[field]["name"]).replace(
            " ", "\\ "
        )
        if self.config.data_field_config[field]["dev_cla"]:
            tags["dev_cla"] = self.config.data_field_config[field]["dev_cla"]
        if self.config.data_field_config[field]["unit"]:
            tags["unit"] = str(self.config.data_field_config[field]["unit"]).replace(
                " ", "\\ "
            )
        tags.update(self.config.data_field_config[field]["tags"])
        return tags.copy()

    def _format_output(self, field, values):
        if field not in self.config.data_field_config:
            return ""
        if not self.config.data_field_config[field]["measurement"]:
            # no measurement, exclude
            return ""

        if self.config.data_field_config[field]["dev_cla"] == "timestamp":
            # skip the epoch fields
            return ""

        asset_class = self.config.data_field_config[field]["asset"]
        attributes = self._get_attributes(values, asset_class)
        nanoepoch = int(
            values[self.timestamp_field.get(asset_class) or "last_update"] * 1000000000
        )

        if field in values and self.config.data_field_config:
            # Get tags
            tags = self._get_tags(field, attributes, asset_class)
            # format output data using the InfluxDB line protocol
            # https://v2.docs.influxdata.com/v2.0/write-data/#line-protocol
            return (
                f'{self.config.data_field_config[field]["measurement"]},'
                f'{",".join("{}={}".format(key, value) for key, value in tags.items())} '
                f"value={values[field]} {nanoepoch}\n"
            )
        else:
            return ""

    def process(self, **args):
        # Send data to influxdb
        try:
            self.access.acquire()
            msg = args["msg"]
            # Log output fields
            self.log_available_fields(msg)

            values = msg.copy()

            # Build structure
            encoded = ""
            for field in values:
                encoded += self._format_output(field, values)

            # Influx has no tables! Use measurement prefix
            # encoded = f'inverter,plant=p1 {",".join("{}={}".format(key, value)
            # for key, value in values.items())} {nanoepoch}'

            # (v1) curl -i -XPOST 'http://localhost:8086/write?db=mydb' --data-binary
            # 'cpu_load_short,host=server01,region=us-west value=0.64 1434055562000000000'
            if self.auth_v2 and self.client:
                with self.client.write_api(write_options=SYNCHRONOUS) as _client:
                    _client.write(bucket=self.bucket, record=encoded)
            else:
                url = f"{'https' if self.ssl else 'http'}://{self.host}:{self.port}/write?db={self.database}"

                r = requests.post(
                    url,
                    data=encoded,
                    headers=self.headers,
                    auth=self.auth,
                    verify=self.verify_ssl,
                )

                r.raise_for_status()

        except InfluxDBError as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"Got error from influxdb2: {e.args} (ignoring: if this happens a lot ... fix it)",
            )

        except requests.exceptions.ConnectionError as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"Got error from influxdb: {e.args} (ignoring: if this happens a lot ... fix it)",
            )

        except requests.exceptions.RequestException as err:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"Unhandled RequestException error: {err}",
            )

        finally:
            self.access.release()



import os
import re
from setuptools import setup, find_packages


def get_version(package):
    """
    Return package version as listed in `__version__` in `__init__.py`.
    """
    path = os.path.join(os.path.dirname(__file__), package, "__init__.py")
    with open(path, "rb") as f:
        init_py = f.read().decode("utf-8")
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


with open("README.md", "r", encoding="UTF-8") as fh:
    long_description = fh.read()

install_requires = [
    "configparser>=3.7.4",
    "requests>=2.21.0",
    "cachetools>=3.1.1",
    "pytz>=2019.1",
    "paho-mqtt>=1.6.1",
    "dsmr-parser>=1.3.2",
    "astral>=1.10.1",
    "pyyaml>=5.1",
    "influxdb-client>=1.26",
]

setup(
    name="omnikdatalogger",
    version=get_version("apps/omnikdatalogger/omnik"),
    license="gpl-3.0",
    author="Jan Bouwhuis",
    author_email="jan@jbsoft.nl",
    description="Omnik Data Logger",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jbouwh/omnikdatalogger",
    package_dir={"": "apps/omnikdatalogger"},
    packages=find_packages(where="apps/omnikdatalogger"),
    data_files=[
        (
            "share/omnikdatalogger",
            [
                "apps/omnikdatalogger/data_fields.json",
                "scripts/systemd/omnikdatalogger.service",
            ],
        )
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Home Automation",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires,
    scripts=["apps/omnikdatalogger/omniklogger.py"],
)
